#!/usr/bin/perl -w
# check_csl - Nagios plugin, which calculates service availability
# level depending on multiple checks' states.
# by Jelena Osadciaja <elenute AT gmail DOT com>

use strict;
use Switch;
use Thread;
use Pod::Usage;
# Change this to path to your Nagios lib folder
use lib "/usr/lib/nagios/plugins/";


use utils qw($TIMEOUT %ERRORS &print_revision &support);
my $PROGNAME = "check_csl";

# Global variables: @data keeps formula's variables name, path to
# executable, variable's value, thread's name, check's output;
# @vars_unique keeps unique formula's variables name, @vars_all-
# all formula's variables names; @warn keeps check's info, which
# returned 'Warning' state, and @crit - which returned 'Critical'
# state. @url will keep URL of instructions page.


my (@data, $formula, @vars_unique, @vars_all, @warn, @crit, @url);

# Option variables
my ($opt_V, $opt_h, $opt_t, $verbose, $opt_hh, $opt_f, $to, $debug, $opt_u);

# Uncomment this  if you don't want to use experimental timeout option
#$SIG{'__WARN__'} = sub{
#    print "unhandeled warning $_[0]";
#};

# Parsing arguments
foreach my $a (@ARGV){
    my $f="-f="; my $ff="--file="; my $t="-t="; my $tt="--timeout=";
    switch ($a) {
	case /^$f/		{ $a=~s/^$f(.*)/$1/;	$opt_f=$a;  }
	case /^$ff/		{ $a=~s/^$ff(.*)/$1/;	$opt_f=$a;  }
	case /^$t/		{ $a=~s/^$t(.*)/$1/; 	$opt_t=$a;  }
	case /^$tt/		{ $a=~s/^$tt(.*)/$1/; 	$opt_t=$a;  }
	case /^-v$|^--verbose$/	{ $verbose++;			    }
	case /^-url$/		{ $opt_u=1;			    }
	case /^-vv$/		{ $verbose+=2;			    }
	case /^-vvv$/		{ $verbose+=3;			    }
	case /^-V$|^--version$/	{ $opt_V=1;			    }
	case /^-h$/		{ $opt_h=1;			    }
	case /^--help$/		{ $opt_hh=1;			    }
	else			{ print "Error: Invalid option: $a\n";
        			  print_usage();
	}
    }
}

if($opt_t){
    if(($opt_t=~/^\d+$/)&&($opt_t>0)) {       $to = $opt_t;
    }else{
            print "Error: Timeout not in seconds!\n";
            exit $ERRORS{'UNKNOWN'};
    }
}else   {
    $to = $TIMEOUT;
}
	
if($opt_V){
    print_revision($PROGNAME,'$Revision: 0.8beta  $ ');
    exit $ERRORS{'OK'};
}

if($verbose){
    $debug=$verbose;
    if($verbose>3)  {   $debug=3;   }
}else{
    $debug=0;
}

if($opt_hh){
    print_help();
    exit $ERRORS{'OK'};
}

if($opt_h){
    print_shorthelp();
    exit $ERRORS{'OK'};
}

if($opt_f){
    if(!(-e $opt_f)){
       print "Error: Configuration file \'$opt_f\' not found!\n";
       exit $ERRORS{'UNKNOWN'};
    }elsif(!(-r $opt_f)){
	print "Error: Configuration file \'$opt_f\' is not readable!\n";
	exit $ERRORS{'UNKNOWN'};
    }else{
	my $res =&performCount($opt_f);
	&print_results($res);
    }
}else{
    print_usage();
}

# Prints debug info accourding to verbosity level
sub debugging{
    my ($msg,$lvl)=@_;
    if($debug>=$lvl){
        print "$msg";
    }
}

# Gets value from formula's variable
sub getValue{
    my $var=$_[0]; my $num=@data;
    for(my $i=0;$i<$num;$i++){
	if($var eq $data[$i][0]){
            return $data[$i][3];
	}
    }
    &debugging("Error: Invalid variable $var.\n",3);
    return 3;
}

# Calculates expression's  service level state
sub countValue{
    my ($a,$b,$o)=@_;
    my $s=$a+$b; my $res=3;
    if($o eq "+"){
        switch ($s) {
            case '0'		{ $res=0 }
            case '4'		{ $res=2 }
            else		{ $res=1 }
	}
    }else{
	switch ($s) {
    	    case '0'		{ $res=0 }
            case '1'		{ $res=1 }
            case [3,4]	     	{ $res=2 }
            else                {
                if(($a==0)||($b==0))	{   $res=2;	}
                else			{   $res=1;	}
            }
	}
    }
    return $res;
}

# Checks if expression has any operator
sub checkForOperator{
    my $exp=$_[0];
    if((index($exp,"+")>=0)||(index($exp,"*")>=0)){
        return 1;
    }else{
        return 0;
    }
}

# If operator is found, splits the expression to two variables
# and operator
sub findOperator {
    my ($o,$exp)=@_;
    my ($start,$p1,$p2,$oo,$str,$isfound);
    $str=$exp; $oo="-"; $isfound=0; $start="";
    while(!$isfound){
	my $ind=index($str,$o);
        if($ind>=0){
            my ($str1,$str2)=&splitString($str,$ind);
	    $start=join("",$start,$str1,$o);
	    if(&checkOperator($start)){
		$p1=substr($start, 0, -1);
		$p2=$str2;
		$oo=$o;
		last;
	    }else{
		$str=$str2;
	    }
        }else{
            last;
	}
    }
    return ($p1,$p2,$oo);
}

# Splits expression string
sub splitString{
    my ($str,$ind)=@_;
    my $exp1=substr($str,0,$ind);
    my $exp2=substr($str,$ind+1,length($str));
    return ($exp1,$exp2);
}

# Checks if operator is 
sub checkOperator{
    my $exp=$_[0];
    my $num1=($exp =~ tr/\(//);
    my $num2=($exp =~ tr/\)//);
    if($num1==$num2)	{  	return 1;    }
    else		{  	return 0;    }
}

# Cuts parentheses if both exist
sub cutParentheses{
    my $exp=$_[0];
    while($exp=~ s/^(\()(.*)(\))$/$2/){};
    return $exp;
}

# Splits the expression into two variables and operator
sub splitExpression{
    my $exp=$_[0];
    $exp=&cutParentheses($exp);
    my ($p1,$p2,$o)=&findOperator("+",$exp);
    if($o eq "-"){
	($p1,$p2,$o)=&findOperator("*",$exp);
    }
    return ($p1,$p2,$o);
}

# Calculates final expression result
sub countFormula{
    my $exp=$_[0];
    my ($p1,$p2,$o)=&splitExpression($exp);
    my ($res1,$res2,$res);
    if($o ne "-"){
	if(&checkForOperator($p1))		{	$res1=&countFormula($p1);	}
	else					{	$res1=&getValue($p1);		}
	if(&checkForOperator($p2))		{  	$res2=&countFormula($p2);	}
	else					{	$res2=&getValue($p2);		}
	    $res=&countValue($res1,$res2,$o);
#	    &debugging("Evaluating: \'$p1\'=\{$res1\} $o \'$p2\'=\{$res2\} = $res\n",3);
   	    return $res;
	}
	elsif(!&checkForOperator($exp)){
	    $res=&getValue($exp);
#	    &debugging("Evaluating: \'$exp\' = $res\n",3);
            return $res;
	}
	else{
	    return -1;
	}
}

# Reads configuration file and populates @data array
sub readFile{
    my $name=$_[0];
    if(open(FILE, "$name" )){
	my $i=0; my $n=0;
	foreach my $line (<FILE>){
	    chomp $line;
	    $line=~s/\s+$//;
	    if($line ne ""){
		my @s=split(/\#/,$line,2);
		$line=$s[0];
		if($line ne ""){
		    my $str="";
		    if(!$i)	{
			if(($opt_u)&&(!$n)){
				$n=1;
				($url[0],$url[1])=split(/ /,$line);
				if((!defined($url[0]))||(!$url[0])){	$url[0]="";	}
				if((!defined($url[1]))||(!$url[1])){	$url[1]=$url[0];}
				$i--;
			}else{	$formula=$line;
			}
		    }else{	($data[$i-1][0],$str)=split(/=/,$line,2);
				if((defined($str))&&($str ne "")&&($str=~/!/)){
				    ($data[$i-1][1],$data[$i-1][2])=split(/ +!/,$str,2);
				     $data[$i-1][2]=~s/\s+$//;
				}else{
				    $data[$i-1][1]=$str;
				    $data[$i-1][2]=$data[$i-1][0];
				}
		                $data[$i-1][3]=-3;
			        if((!defined($data[$i-1][2]))||($data[$i-1][2] eq "")){
				    $data[$i-1][2]=$data[$i-1][0];
		    	       }
		    }
		$i++;
		}
	    }
        }
	close(FILE);
	return 1;
    }else{
	return 0;
    }
}

# Checks formula for errors
sub checkFormula{
    my $conf=$_[0];
    if(!&checkForWhiteSpace)		{ &debugging("Error: Formula \'$formula\' has white space\n",2);                 }
    elsif(!&checkNumOfParentheses) 	{ &debugging("Error: Formula \'$formula\' has odd number of parentheses\n",2)    }
    elsif(!&checkNumOfVars) 		{ &debugging("Error: Formula \'$formula\' has invalid number of variables\n",2)  }
    elsif(!&checkNumOfOperators)	{ &debugging("Error: Formula \'$formula\' has invalid number of operators\n",2)  }
    elsif(!&checkForTypos)		{ &debugging("Error: Formula \'$formula\' has typing errors\n",2)                }
    else                                { return 1;                                                         }
    &debugging("Error: Configuration file \'$conf\'  has errors in formula\n",0);
    return 0;
}

# Checks if configuration file has valid number of variables defined
sub checkNumOfVars{
    my $num=@data;
    my $numf=&checkNumOfVarsForm;
    if($num==$numf)	{	return 1;	}
    else		{	return 0;	}
}

# Populates the @vars_unique and @vars_all arrays
sub checkNumOfVarsForm{                    
    my $f=$formula;
    while($f =~ /(\w+)/g) {
	if(!grep(/$1/, @vars_unique)){
            push(@vars_unique,$1);
        }
	push(@vars_all,$1);
    }
    return @vars_unique;
}

# Checks if formula has odd number of parentheses
sub checkNumOfParentheses{
    my $f=$formula;
    if(&checkOperator($f))	{	return 1;       }
    else			{	return 0;	}
}

# Checks if formula has valid number of operators
sub checkNumOfOperators{
    my $num=($formula =~ tr/\*|\+//);
    if($num+1==@vars_all)	{  	return 1;       }
    else			{  	return 0;       }
}

# Checks if formula has white spaces
sub checkForWhiteSpace{
    if($formula=~/\s/)	        {	return 0;	}
    else		        {	return 1;	}
}

# Checks for other typing errors
sub checkForTypos{
    if($formula=~/\(\)|\)\(|\*{2,}|\*{2,}|(\*\+)+|(\+\*)+|^[^\w\)\(]|[^\w\)\(]$/){
	return 0;
    }else{
	return 1;
    }
}

# Checks if Check's section has errors
sub checkExecutables{
    my $conf=$_[0];
    for(my $i=0;$i<@data;$i++){
	my $line=$data[$i][1];
	my @file=split(/ /,$line);
	if(!(-e $file[0])){
            &debugging("Error: Check \'$file[0]\' not found\n",2);
            &debugging("Error: Configuration file \'$conf\' has errors in check's description\n",0);
            return 0;
	}
    }
    return 1;
}

# Returns final service level state
sub performCount{
    my $file=$_[0];
    if(&readFile($file)){
   	if((!&checkFormula($file))||(!&checkExecutables($file))||(!&checkDescription($file))){
            exit $ERRORS{'UNKNOWN'};
        }else{
            if(&performChecks){
		my $res=&countFormula($formula);
                return $res;
	    }else{
		&debugging("Error performing checks\n",0);
                exit $ERRORS{'UNKNOWN'};
	    }
        }
    }else{
	&debugging("Error: Configuration file \'$file\' is not readable\n",0);
        exit $ERRORS{'UNKNOWN'};
    }
}

# Runs check's executable file with arguments
sub runCheck{
    my $args = ($_[0]);
    my @output =`$args 2>&1`;
    my $res=$? >> 8;
    if($res>128)    {   $res-=256;  };
    return ($res, @output);
}

# Runs checks in threads and populates @data array
sub performChecks{
    my $err=0; my $name;
    for(my $i=0;$i<@data;$i++){
	($data[$i][4])=Thread->new(\&runCheck,$data[$i][1]);
    }
    my $t=time();
    for(my $i=0;$i<@data;$i++){
	if(time()<$t+$to){
            ($data[$i][3],$data[$i][5])=$data[$i][4]->join;
	    if(($data[$i][3]<0)||($data[$i][3]>2)){
		$err=1;
                &debugging("Error: Check \"$data[$i][2]\" ($data[$i][1]) exited with state \'UNKNOWN\'\n",2);
		&debugging("$data[$i][5]\n",3);
		return 0;
	    }
	    if($data[$i][3]==1)		{	push(@warn,$data[$i][2]);	}
	    elsif($data[$i][3]==2)	{	push(@crit,$data[$i][2]);	}
        }else{
            &debugging("Error: Timeout $to.s exeeded\n",0);
            $err=1;
            last;
	}
    }
    if(!$err){
	return 1;
    }else{
	for(my $i=0;$i<@data;$i++){
	    $data[$i][4]->detach();
        }
        return 0;
    }
}

# Checks if formula has all variables descriptions
sub checkDescription{
    my $conf=$_[0];
    my $num=@data;
    for(my $ind=0;$ind<@data;$ind++){
	if(!grep(/$data[$ind][0]/,@vars_unique)){
	    &debugging("Error: Check's \'$data[$ind][2]\' description is invalid\n",3);
	    &debugging("Error: Configuration file \'$conf\' has errors in check's description\n",0);
	    return 0;
	}
    }
    return 1;
}

# Prints usage info
sub print_usage{
    print "Usage: $PROGNAME.pl -f=<file> [-t=<timeout>] [-v|-vv|-vvv] [-url] [-h] [--help] [-V]\n\n";
    support();
    exit $ERRORS{"UNKNOWN"};
}

# Prints extended help
sub print_help{
    pod2usage(  -verbose => '2');
}

# Prints short help
sub print_shorthelp{
    pod2usage(  -verbose => '1');
}

# Prints results according to debug level
sub print_results{
    my $res=$_[0];
    my (@cw,$line,$line1,$line2,$c,$cw,$cc,$state,$msg,$dot);
    if($opt_u){
	$dot="<a href=\"".$url[0]."\">".$url[1]."</a>";
    }else{
	$dot="";
    }
    switch($res){
	    case '0' {	$state="OK"; 		}
	    case '1' {	$state="WARNING";	}
	    case '2' {	$state="CRITICAL";	}
    }
    if($res==0) 	{ print "$state.\n";
    }elsif($debug==0)	{ print "$state. $dot\n"; 
    }else{
	$line1="";$line2=""; $c="Check(s)";$cw="returned state \'WARNING\'";$cc="returned state \'CRITICAL\'";
	if(@warn>0){
	    foreach my $q(@warn)	{	$line1=$line1." \'$q\',";	}
	    $line1=~s/,$//;			$line1=$c.$line1." ".$cw.".";
	}
	if(@crit>0){
	    foreach my $q (@crit)	{	$line2=$line2." \'$q\',";	}
	    $line2=~s/,$//;			$line2=$c.$line2." ".$cc.".";
	}
	$line=$line1." ".$line2;
	if($debug==1) 			{	print "$state: $line $dot\n";		}
	else{
	    print "$state: $dot\n";
	    &print_results_deb($cw,@warn);
	    &print_results_deb($cc,@crit);
	}
    }
    exit $ERRORS{$state};
}

# Prints extended output
sub print_results_deb{
    my($message,@array)=@_;
    if(@array >0){
	for(my $i=0;$i<@array;$i++){
	    for(my $j=0;$j<@data;$j++){
	        if($data[$j][2] eq $array[$i]){
		    if($debug==2){
		        print "Check \'$data[$j][2]\' ($data[$j][1]) $message.\n";
		    }else{
		        print "Check \'$data[$j][2]\' ($data[$j][1]) $message with output:\n";
		        if((defined($data[$j][5]))&&($data[$j][5]ne ""))	{	
											my @a=$data[$j][5];
										        chomp $a[$#a];
											print "  @a\n";
		        }else							{	print "No output produced.\n";}
		    }
		}
	    }
	}
    }
}
__END__
 
=head1 NAME

check_csl is Nagios plug-in, which calculates service availability level
depending on multiple checks' states.

=head1 VERSION

This document describes version 0.8 beta of check_csl, released March 17, 2006.

=head1 SYNOPSIS

    check_csl.pl option=value [switch(es)]
        Available options are:
        	-f,--file	Configuration file name - required;
        	-t,--timeout	Plug-in timeout in seconds;
        Available switches are:
		-url		Provide instructions URL
        	-V,--version	Plug-in version;
        	-v,--verbose	Verbosity level (up to 3);
        	-h		Usage help;
                   --help	Manual help.

=head1 DESCRIPTION

check_csl runs necessary checks and according to defined conditions,
interprets results and returns service level -  C<OK>, C<Warning> or
C<Critical>.

Calculation conditions are written to configuration file in form of formula.
Also information about checks, executed by plug-in, is included in same
configuration file. It is also possible to specify custom plug-in runtime
timeout and desirable verbosity level.

=head2 Configuration File

Configuration file consists of three sections: Instruction section,formula
section, according which the final result will be calculated, and checks
description section, where checks executables are written. Comments, beginning
with comment sign C<#> are allowed.

=head3 Instructions URL Section

Instructions section is only required when C<-url> switch is enabled. This section
must begin whith first uncommented configuration file line. The URL with instructions
will be given with C<WARNING> and C<CRITICAL> output. The sections contains one
line with URL and short URL descritpion (not required), separated by white
space. If I<no> C<-url> switch specified, this line MUST be commented.

=head3 Formula Section

If no C<-url> switch specified,Formula section must begin from first uncommented
configuration file line, else, the section must begin with second uncommented file
line. Formula contains expression, which may consist of variables, operators and
parentheses. Variable keeps state of a particular check, operators describe
the way results will be interpreted, and parentheses change the expression
execution priority.

=head4 Variables

Check variable keeps calculated check's state. Every different check must have
one and only variable. Information about what check variable is related to is
kept in Check's description section. Name of the variable must contain only
"word" symbols (alphanumeric or C<_>), no white spaces are allowed. Names are
case-sensitive. Variables are separated between each other with operator sign
or parentheses too. Variables keep C<OK>, C<Warning> or C<Critical> values.

=head4 Operators

Within calculations two types of operators are used: I<logical OR>  marked
with C<+> sign and I<logical AND> marked with C<*> sign. I<logical AND> operator
has higher priority level than I<logical OR> operator. Execution priority may be
changed with parentheses.

If expression contains two variables and operator sign, the calculated
result would be:

    1 var's value: OK   OK	 OK	     Warning    Warning	    Critical
    2 var's value: OK   Warning  Critical    Warning	Critical    Critical

    Result of "+": OK   Warning	 Warning     Warning	Warning	    Critical
    Result of "*": OK   Warning	 Critical    Warning	Critical    Critical

=head4 Parentheses

Parentheses change the order of operations. Covered expression will be
evaluated first. If there are nested parentheses, the expression covered with
inner parentheses is evaluated first. Overall:

=over 4

=item *

If expression has parentheses, the covered expression is evaluated first.

=item *

Result of I<logical AND> is calculated.

=item *

Result of I<logical OR> is calculated.

=back

=head3 Checks Section

In the Check's description section the check's variable name, check's full
path to the executable, and short description (not required), are written.
This section must begin with first uncommented line after the Formula section.

=head4 Check's variable

Every check variable is written on the separate line. Requirements for the
variable names are written in Variables section.

=head4 Check's executable

After the variable's name and equation sign C<=> follows the path to check's
file and its arguments. Nagios must have proper permissions to execute check.

=head4 Check's description

Since availability level calculations may include same check's names with
different arguments, the use of check's file names for short description is
not convenient. Also, using the variable's name for short check description
may not be convenient either. Therefore explicit short check description is
allowed after the exclamation mark C<!>. It is possible to specify distinct
variable name or include a short check description.

=head2 Example of Configuration File
    
    ###################################################################################
    #										      #
    #				Start of Instrucrions URL section		      #
    http://www.google.com instructions
    #										      #
    #				End of Instructions URL section			      #
    #		(Comment this section if no "-url" swith will be used)		      #
    ###################################################################################
    #										      #
    #				Start of Formula's section			      #
    (a+b)*host3										
    #										      #
    #                           End of Formula's section			      #
    ###################################################################################
    #										      #
    #				Start of Check's Description section		      #
    #										      #
    a=/usr/share/nagios/plugins/check_ping -H google.lt -w 100,10%% -c 200,20%% !google
    b=/usr/share/nagios/plugins/check_ping -H cisco.com -w 150,30%% -c 400,80%% !cisco
    host3=/usr/share/nagios/plugins/check_ping -H 192.168.0.1 -w 30,5%% -c 50,50%%
    # 										      #
    #				End of Check's Description section		      #
    #										      #
    # The expression above shows, that the availability level can be achieved if right# 
    # host3 availability level and one of the google or cisco availability levels     #
    # condition met.								      #
    ###################################################################################

=head2 Runtime Timeout

All plug-ins have limited runtime, and the default timeout value is specified
in the Nagios configuration file. It is possible to specify explicit timeout
value in seconds with option C<--timeout>. Since every check is run in separate
thread, when plug-in timeouts because of timeouted check, it detaches all
threads, leaving them work by their own until they're done, and exits with
state C<UNKNOWN>. This feature is highly experimental.

=head2 Instructions URL

If this switch is enabled, with C<WARNING> or C<CRITICAL> output the additional
HTML coded URL will be given. The page, URL is pointing to, should contain some
instructions for the contact groups. The URL and short description must be written
in configuration file, that plug-in could return 
C<E<lt>a href="url"E<gt>short_descriptionE<lt>/aE<gt>>.

=head2 Verbose Level

With default verbosity level plug-in produces single line output: C<Ok>,
C<Warning> or C<Critical>.

With first verbose level plug-in returns single line output with additional
information: short check's description, which returned state C<Warning> or
C<Critical>.

With second verbose level plug-in produces multi line output,
specifying on each separate line which checks returned state C<Warning> and
C<Critical> and what arguments were they called with.

If plug-in is run under third verbosity level, it generates multi line output,
additionally including check's generated output.

=head1 AUTHOR

Jelena Osadciaja (elenute AT gmail DOT com).

=head1 COPYRIGHT AND LICENCE

check_csl is licensed under the terms of the GNU General Public License Version
2 as published by the Free Software Foundation.

Like Nagios itself, check_csl is provided AS IS with NO WARRANTY OF ANY KIND,
INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE.
